@@ -6,69 +6,53 @@
 # http://svn.apache.org/viewvc/avro/trunk/lang/py/src/avro/ which is under
 # Apache 2.0 license (http://www.apache.org/licenses/LICENSE-2.0)
 
-from io import BytesIO
-from struct import error as StructError
 import bz2
+import json
 import lzma
 import zlib
-from datetime import datetime, time, date, timezone, timedelta
+from datetime import datetime, timezone
 from decimal import Context
-from uuid import UUID
-
-import json
+from io import BytesIO
+from struct import error as StructError
+from typing import IO, Union, Optional, Generic, TypeVar, Iterator, Dict
 
 from .io.binary_decoder import BinaryDecoder
 from .io.json_decoder import AvroJSONDecoder
-from .schema import extract_record_type, extract_logical_type, parse_schema
+from .logical_readers import LOGICAL_READERS
+from .schema import (
+    extract_record_type,
+    is_single_record_union,
+    is_single_name_union,
+    extract_logical_type,
+    parse_schema,
+)
+from .types import Schema, AvroMessage, NamedSchemas
 from ._read_common import (
     SchemaResolutionError,
     MAGIC,
     SYNC_SIZE,
     HEADER_SCHEMA,
     missing_codec_lib,
-    NAMED_TYPES,
-)
-from .const import (
-    MCS_PER_HOUR,
-    MCS_PER_MINUTE,
-    MCS_PER_SECOND,
-    MLS_PER_HOUR,
-    MLS_PER_MINUTE,
-    MLS_PER_SECOND,
-    DAYS_SHIFT,
 )
+from .const import NAMED_TYPES, AVRO_TYPES
 
-MASK = 0xFF
-AVRO_TYPES = {
-    "boolean",
-    "bytes",
-    "double",
-    "float",
-    "int",
-    "long",
-    "null",
-    "string",
-    "fixed",
-    "enum",
-    "record",
-    "error",
-    "array",
-    "map",
-    "union",
-    "request",
-    "error_union",
-}
+T = TypeVar("T")
 
 decimal_context = Context()
 epoch = datetime(1970, 1, 1, tzinfo=timezone.utc)
+epoch_naive = datetime(1970, 1, 1)
 
 
-def match_types(writer_type, reader_type):
+def _default_named_schemas() -> Dict[str, NamedSchemas]:
+    return {"writer": {}, "reader": {}}
+
+
+def match_types(writer_type, reader_type, named_schemas):
     if isinstance(writer_type, list) or isinstance(reader_type, list):
         return True
     if isinstance(writer_type, dict) or isinstance(reader_type, dict):
         try:
-            return match_schemas(writer_type, reader_type)
+            return match_schemas(writer_type, reader_type, named_schemas)
         except SchemaResolutionError:
             return False
     if writer_type == reader_type:
@@ -84,10 +68,14 @@ def match_types(writer_type, reader_type):
         return True
     elif writer_type == "bytes" and reader_type == "string":
         return True
+    writer_schema = named_schemas["writer"].get(writer_type)
+    reader_schema = named_schemas["reader"].get(reader_type)
+    if writer_schema is not None and reader_schema is not None:
+        return match_types(writer_schema, reader_schema, named_schemas)
     return False
 
 
-def match_schemas(w_schema, r_schema):
+def match_schemas(w_schema, r_schema, named_schemas):
     error_msg = f"Schema mismatch: {w_schema} is not {r_schema}"
     if isinstance(w_schema, list):
         # If the writer is a union, checks will happen in read_union after the
@@ -97,7 +85,7 @@ def match_schemas(w_schema, r_schema):
         # If the reader is a union, ensure one of the new schemas is the same
         # as the writer
         for schema in r_schema:
-            if match_types(w_schema, schema):
+            if match_types(w_schema, schema, named_schemas):
                 return schema
         else:
             raise SchemaResolutionError(error_msg)
@@ -113,17 +101,26 @@ def match_schemas(w_schema, r_schema):
             r_type = r_schema
 
         if w_type == r_type == "map":
-            if match_types(w_schema["values"], r_schema["values"]):
+            if match_types(w_schema["values"], r_schema["values"], named_schemas):
                 return r_schema
         elif w_type == r_type == "array":
-            if match_types(w_schema["items"], r_schema["items"]):
+            if match_types(w_schema["items"], r_schema["items"], named_schemas):
                 return r_schema
         elif w_type in NAMED_TYPES and r_type in NAMED_TYPES:
-            if w_schema["name"] == r_schema["name"] or w_schema["name"] in r_schema.get(
-                "aliases", []
-            ):
+            if w_type == r_type == "fixed" and w_schema["size"] != r_schema["size"]:
+                raise SchemaResolutionError(
+                    f"Schema mismatch: {w_schema} size is different than {r_schema} size"
+                )
+
+            w_unqual_name = w_schema["name"].split(".")[-1]
+            r_unqual_name = r_schema["name"].split(".")[-1]
+            r_aliases = r_schema.get("aliases", [])
+            if w_unqual_name == r_unqual_name or w_schema["name"] in r_aliases:
                 return r_schema
-        elif match_types(w_type, r_type):
+        elif w_type not in AVRO_TYPES and r_type in NAMED_TYPES:
+            if match_types(w_type, r_schema["name"], named_schemas):
+                return r_schema["name"]
+        elif match_types(w_type, r_type, named_schemas):
             return r_schema
         raise SchemaResolutionError(error_msg)
 
@@ -133,7 +130,7 @@ def read_null(
     writer_schema=None,
     named_schemas=None,
     reader_schema=None,
-    return_record_name=False,
+    options={},
 ):
     return decoder.read_null()
 
@@ -147,7 +144,7 @@ def read_boolean(
     writer_schema=None,
     named_schemas=None,
     reader_schema=None,
-    return_record_name=False,
+    options={},
 ):
     return decoder.read_boolean()
 
@@ -156,58 +153,12 @@ def skip_boolean(decoder, writer_schema=None, named_schemas=None):
     decoder.read_boolean()
 
 
-def read_timestamp_millis(data, writer_schema=None, reader_schema=None):
-    # Cannot use datetime.fromtimestamp: https://bugs.python.org/issue36439
-    return epoch + timedelta(microseconds=data * 1000)
-
-
-def read_timestamp_micros(data, writer_schema=None, reader_schema=None):
-    # Cannot use datetime.fromtimestamp: https://bugs.python.org/issue36439
-    return epoch + timedelta(microseconds=data)
-
-
-def read_date(data, writer_schema=None, reader_schema=None):
-    return date.fromordinal(data + DAYS_SHIFT)
-
-
-def read_uuid(data, writer_schema=None, reader_schema=None):
-    return UUID(data)
-
-
-def read_time_millis(data, writer_schema=None, reader_schema=None):
-    h = int(data / MLS_PER_HOUR)
-    m = int(data / MLS_PER_MINUTE) % 60
-    s = int(data / MLS_PER_SECOND) % 60
-    mls = int(data % MLS_PER_SECOND) * 1000
-    return time(h, m, s, mls)
-
-
-def read_time_micros(data, writer_schema=None, reader_schema=None):
-    h = int(data / MCS_PER_HOUR)
-    m = int(data / MCS_PER_MINUTE) % 60
-    s = int(data / MCS_PER_SECOND) % 60
-    mcs = data % MCS_PER_SECOND
-    return time(h, m, s, mcs)
-
-
-def read_decimal(data, writer_schema=None, reader_schema=None):
-    scale = writer_schema.get("scale", 0)
-    precision = writer_schema["precision"]
-
-    unscaled_datum = int.from_bytes(data, byteorder="big", signed=True)
-
-    decimal_context.prec = precision
-    return decimal_context.create_decimal(unscaled_datum).scaleb(
-        -scale, decimal_context
-    )
-
-
 def read_int(
     decoder,
     writer_schema=None,
     named_schemas=None,
     reader_schema=None,
-    return_record_name=False,
+    options={},
 ):
     return decoder.read_int()
 
@@ -221,7 +172,7 @@ def read_long(
     writer_schema=None,
     named_schemas=None,
     reader_schema=None,
-    return_record_name=False,
+    options={},
 ):
     return decoder.read_long()
 
@@ -235,7 +186,7 @@ def read_float(
     writer_schema=None,
     named_schemas=None,
     reader_schema=None,
-    return_record_name=False,
+    options={},
 ):
     return decoder.read_float()
 
@@ -249,7 +200,7 @@ def read_double(
     writer_schema=None,
     named_schemas=None,
     reader_schema=None,
-    return_record_name=False,
+    options={},
 ):
     return decoder.read_double()
 
@@ -263,7 +214,7 @@ def read_bytes(
     writer_schema=None,
     named_schemas=None,
     reader_schema=None,
-    return_record_name=False,
+    options={},
 ):
     return decoder.read_bytes()
 
@@ -277,9 +228,11 @@ def read_utf8(
     writer_schema=None,
     named_schemas=None,
     reader_schema=None,
-    return_record_name=False,
+    options={},
 ):
-    return decoder.read_utf8()
+    return decoder.read_utf8(
+        handle_unicode_errors=options.get("handle_unicode_errors", "strict")
+    )
 
 
 def skip_utf8(decoder, writer_schema=None, named_schemas=None):
@@ -291,7 +244,7 @@ def read_fixed(
     writer_schema,
     named_schemas=None,
     reader_schema=None,
-    return_record_name=False,
+    options={},
 ):
     size = writer_schema["size"]
     return decoder.read_fixed(size)
@@ -307,7 +260,7 @@ def read_enum(
     writer_schema,
     named_schemas,
     reader_schema=None,
-    return_record_name=False,
+    options={},
 ):
     symbol = writer_schema["symbols"][decoder.read_enum()]
     if reader_schema and symbol not in reader_schema["symbols"]:
@@ -330,28 +283,28 @@ def read_array(
     writer_schema,
     named_schemas,
     reader_schema=None,
-    return_record_name=False,
+    options={},
 ):
     if reader_schema:
 
-        def item_reader(decoder, w_schema, r_schema, return_record_name):
+        def item_reader(decoder, w_schema, r_schema, options):
             return read_data(
                 decoder,
                 w_schema["items"],
                 named_schemas,
                 r_schema["items"],
-                return_record_name,
+                options,
             )
 
     else:
 
-        def item_reader(decoder, w_schema, _, return_record_name):
+        def item_reader(decoder, w_schema, r_schema, options):
             return read_data(
                 decoder,
                 w_schema["items"],
                 named_schemas,
                 None,
-                return_record_name,
+                options,
             )
 
     read_items = []
@@ -360,7 +313,12 @@ def read_array(
 
     for item in decoder.iter_array():
         read_items.append(
-            item_reader(decoder, writer_schema, reader_schema, return_record_name)
+            item_reader(
+                decoder,
+                writer_schema,
+                reader_schema,
+                options,
+            )
         )
 
     decoder.read_array_end()
@@ -382,7 +340,7 @@ def read_map(
     writer_schema,
     named_schemas,
     reader_schema=None,
-    return_record_name=False,
+    options={},
 ):
     if reader_schema:
 
@@ -392,18 +350,18 @@ def read_map(
                 w_schema["values"],
                 named_schemas,
                 r_schema["values"],
-                return_record_name,
+                options,
             )
 
     else:
 
-        def item_reader(decoder, w_schema, _):
+        def item_reader(decoder, w_schema, r_schema):
             return read_data(
                 decoder,
                 w_schema["values"],
                 named_schemas,
                 None,
-                return_record_name,
+                options,
             )
 
     read_items = {}
@@ -434,7 +392,7 @@ def read_union(
     writer_schema,
     named_schemas,
     reader_schema=None,
-    return_record_name=False,
+    options={},
 ):
     # schema resolution
     index = decoder.read_index()
@@ -443,46 +401,49 @@ def read_union(
     if reader_schema:
         # Handle case where the reader schema is just a single type (not union)
         if not isinstance(reader_schema, list):
-            if match_types(idx_schema, reader_schema):
+            if match_types(idx_schema, reader_schema, named_schemas):
                 return read_data(
                     decoder,
                     idx_schema,
                     named_schemas,
                     reader_schema,
-                    return_record_name,
+                    options,
                 )
         else:
             for schema in reader_schema:
-                if match_types(idx_schema, schema):
+                if match_types(idx_schema, schema, named_schemas):
                     return read_data(
                         decoder,
                         idx_schema,
                         named_schemas,
                         schema,
-                        return_record_name,
+                        options,
                     )
         msg = f"schema mismatch: {writer_schema} not found in {reader_schema}"
         raise SchemaResolutionError(msg)
     else:
-        if return_record_name and extract_record_type(idx_schema) == "record":
-            return (
-                idx_schema["name"],
-                read_data(
-                    decoder,
-                    idx_schema,
-                    named_schemas,
-                    None,
-                    return_record_name,
-                ),
-            )
+        result = read_data(decoder, idx_schema, named_schemas, None, options)
+
+        return_record_name_override = options.get("return_record_name_override")
+        return_record_name = options.get("return_record_name")
+        return_named_type_override = options.get("return_named_type_override")
+        return_named_type = options.get("return_named_type")
+        if return_named_type_override and is_single_name_union(writer_schema):
+            return result
+        elif return_named_type and extract_record_type(idx_schema) in NAMED_TYPES:
+            return (idx_schema["name"], result)
+        elif return_named_type and extract_record_type(idx_schema) not in AVRO_TYPES:
+            # idx_schema is a named type
+            return (named_schemas["writer"][idx_schema]["name"], result)
+        elif return_record_name_override and is_single_record_union(writer_schema):
+            return result
+        elif return_record_name and extract_record_type(idx_schema) == "record":
+            return (idx_schema["name"], result)
         elif return_record_name and extract_record_type(idx_schema) not in AVRO_TYPES:
             # idx_schema is a named type
-            return (
-                named_schemas[idx_schema]["name"],
-                read_data(decoder, idx_schema, named_schemas, None, return_record_name),
-            )
+            return (named_schemas["writer"][idx_schema]["name"], result)
         else:
-            return read_data(decoder, idx_schema, named_schemas)
+            return result
 
 
 def skip_union(decoder, writer_schema, named_schemas):
@@ -496,7 +457,7 @@ def read_record(
     writer_schema,
     named_schemas,
     reader_schema=None,
-    return_record_name=False,
+    options={},
 ):
     """A record is encoded by encoding the values of its fields in the order
     that they are declared. In other words, a record is encoded as just the
@@ -520,7 +481,11 @@ def read_record(
     if reader_schema is None:
         for field in writer_schema["fields"]:
             record[field["name"]] = read_data(
-                decoder, field["type"], named_schemas, None, return_record_name
+                decoder,
+                field["type"],
+                named_schemas,
+                None,
+                options,
             )
     else:
         readers_field_dict = {}
@@ -541,7 +506,7 @@ def read_record(
                     field["type"],
                     named_schemas,
                     readers_field["type"],
-                    return_record_name,
+                    options,
                 )
             else:
                 skip_data(decoder, field["type"], named_schemas)
@@ -565,17 +530,6 @@ def skip_record(decoder, writer_schema, named_schemas):
         skip_data(decoder, field["type"], named_schemas)
 
 
-LOGICAL_READERS = {
-    "long-timestamp-millis": read_timestamp_millis,
-    "long-timestamp-micros": read_timestamp_micros,
-    "int-date": read_date,
-    "bytes-decimal": read_decimal,
-    "fixed-decimal": read_decimal,
-    "string-uuid": read_uuid,
-    "int-time-millis": read_time_millis,
-    "long-time-micros": read_time_micros,
-}
-
 READERS = {
     "null": read_null,
     "boolean": read_boolean,
@@ -633,19 +587,22 @@ def maybe_promote(data, writer_type, reader_type):
 
 
 def read_data(
-    decoder, writer_schema, named_schemas, reader_schema=None, return_record_name=False
+    decoder,
+    writer_schema,
+    named_schemas,
+    reader_schema=None,
+    options={},
 ):
     """Read data from file object according to schema."""
 
     record_type = extract_record_type(writer_schema)
 
-    if reader_schema and record_type in AVRO_TYPES:
-        # If the schemas are the same, set the reader schema to None so that no
-        # schema resolution is done for this call or future recursive calls
-        if writer_schema == reader_schema:
-            reader_schema = None
-        else:
-            reader_schema = match_schemas(writer_schema, reader_schema)
+    if reader_schema:
+        reader_schema = match_schemas(
+            writer_schema,
+            reader_schema,
+            named_schemas,
+        )
 
     reader_fn = READERS.get(record_type)
     if reader_fn:
@@ -655,7 +612,7 @@ def read_data(
                 writer_schema,
                 named_schemas,
                 reader_schema,
-                return_record_name,
+                options,
             )
         except StructError:
             raise EOFError(f"cannot read {record_type} from {decoder.fo}")
@@ -673,10 +630,10 @@ def read_data(
     else:
         return read_data(
             decoder,
-            named_schemas[record_type],
+            named_schemas["writer"][record_type],
             named_schemas,
-            named_schemas.get(reader_schema),
-            return_record_name,
+            named_schemas["reader"].get(reader_schema),
+            options,
         )
 
 
@@ -687,7 +644,7 @@ def skip_data(decoder, writer_schema, named_schemas):
     if reader_fn:
         reader_fn(decoder, writer_schema, named_schemas)
     else:
-        skip_data(decoder, named_schemas[record_type], named_schemas)
+        skip_data(decoder, named_schemas["writer"][record_type], named_schemas)
 
 
 def skip_sync(fo, sync_marker):
@@ -706,7 +663,7 @@ def deflate_read_block(decoder):
     data = decoder.read_bytes()
     # -15 is the log of the window size; negative indicates "raw" (no
     # zlib headers) decompression.  See zlib.h.
-    return BytesIO(zlib.decompress(data, -15))
+    return BytesIO(zlib.decompressobj(-15).decompress(data))
 
 
 def bzip2_read_block(decoder):
@@ -747,7 +704,7 @@ else:
 def zstandard_read_block(decoder):
     length = read_long(decoder)
     data = decoder.read_fixed(length)
-    return BytesIO(zstd.ZstdDecompressor().decompress(data))
+    return BytesIO(zstd.ZstdDecompressor().decompressobj().decompress(data))
 
 
 try:
@@ -779,7 +736,7 @@ def _iter_avro_records(
     writer_schema,
     named_schemas,
     reader_schema,
-    return_record_name=False,
+    options,
 ):
     """Return iterator over avro records."""
     sync_marker = header["sync"]
@@ -803,7 +760,7 @@ def _iter_avro_records(
                 writer_schema,
                 named_schemas,
                 reader_schema,
-                return_record_name,
+                options,
             )
 
         skip_sync(decoder.fo, sync_marker)
@@ -816,7 +773,7 @@ def _iter_avro_blocks(
     writer_schema,
     named_schemas,
     reader_schema,
-    return_record_name=False,
+    options,
 ):
     """Return iterator over avro blocks."""
     sync_marker = header["sync"]
@@ -847,7 +804,7 @@ def _iter_avro_blocks(
             named_schemas,
             offset,
             size,
-            return_record_name,
+            options,
         )
 
 
@@ -868,7 +825,7 @@ class Block:
 
     .. attribute:: offset
 
-        Offset of the block from the begining of the avro file
+        Offset of the block from the beginning of the avro file
 
     .. attribute:: size
 
@@ -885,7 +842,7 @@ class Block:
         named_schemas,
         offset,
         size,
-        return_record_name=False,
+        options,
     ):
         self.bytes_ = bytes_
         self.num_records = num_records
@@ -895,7 +852,7 @@ class Block:
         self._named_schemas = named_schemas
         self.offset = offset
         self.size = size
-        self.return_record_name = return_record_name
+        self.options = options
 
     def __iter__(self):
         for i in range(self.num_records):
@@ -904,7 +861,7 @@ class Block:
                 self.writer_schema,
                 self._named_schemas,
                 self.reader_schema,
-                self.return_record_name,
+                self.options,
             )
 
     def __str__(self):
@@ -915,24 +872,28 @@ class Block:
         )
 
 
-class file_reader:
-    def __init__(self, fo_or_decoder, reader_schema=None, return_record_name=False):
+class file_reader(Generic[T]):
+    def __init__(
+        self,
+        fo_or_decoder,
+        reader_schema=None,
+        options={},
+    ):
         if isinstance(fo_or_decoder, AvroJSONDecoder):
             self.decoder = fo_or_decoder
         else:
             # If a decoder was not provided, assume binary
             self.decoder = BinaryDecoder(fo_or_decoder)
 
-        self._named_schemas = {}
+        self._named_schemas = _default_named_schemas()
         if reader_schema:
             self.reader_schema = parse_schema(
-                reader_schema,
-                _write_hint=False,
-                _named_schemas=self._named_schemas,
+                reader_schema, self._named_schemas["reader"], _write_hint=False
             )
+
         else:
             self.reader_schema = None
-        self.return_record_name = return_record_name
+        self.options = options
         self._elems = None
 
     def _read_header(self):
@@ -942,9 +903,9 @@ class file_reader:
                 HEADER_SCHEMA,
                 self._named_schemas,
                 None,
-                self.return_record_name,
+                self.options,
             )
-        except StopIteration:
+        except (StopIteration, EOFError):
             raise ValueError("cannot read header - is it an avro file?")
 
         # `meta` values are bytes. So, the actual decoding has to be external.
@@ -953,13 +914,24 @@ class file_reader:
         self._schema = json.loads(self.metadata["avro.schema"])
         self.codec = self.metadata.get("avro.codec", "null")
 
+        # Older avro files created before we were more strict about
+        # defaults might have been writen with a bad default. Since we re-parse
+        # the writer schema here, it will now fail. Therefore, if a user
+        # provides a reader schema that passes parsing, we will ignore those
+        # default errors
+        if self.reader_schema is not None:
+            ignore_default_error = True
+        else:
+            ignore_default_error = False
+
         # Always parse the writer schema since it might have named types that
         # need to be stored in self._named_types
         self.writer_schema = parse_schema(
             self._schema,
+            self._named_schemas["writer"],
             _write_hint=False,
             _force=True,
-            _named_schemas=self._named_schemas,
+            _ignore_default_error=ignore_default_error,
         )
 
     @property
@@ -972,30 +944,49 @@ class file_reader:
         )
         return self._schema
 
-    def __iter__(self):
+    def __iter__(self) -> Iterator[T]:
         if not self._elems:
             raise NotImplementedError
         return self._elems
 
-    def next(self):
+    def __next__(self) -> T:
         return next(self._elems)
 
-    __next__ = next
 
-
-class reader(file_reader):
+class reader(file_reader[AvroMessage]):
     """Iterator over records in an avro file.
 
     Parameters
     ----------
-    fo: file-like
-        Input stream
-    reader_schema: dict, optional
+    fo
+        File-like object to read from
+    reader_schema
         Reader schema
-    return_record_name: bool, optional
+    return_record_name
         If true, when reading a union of records, the result will be a tuple
         where the first value is the name of the record and the second value is
         the record itself
+    return_record_name_override
+        If true, this will modify the behavior of return_record_name so that
+        the record name is only returned for unions where there is more than
+        one record. For unions that only have one record, this option will make
+        it so that the record is returned by itself, not a tuple with the name.
+    return_named_type
+        If true, when reading a union of named types, the result will be a tuple
+        where the first value is the name of the type and the second value is
+        the record itself
+        NOTE: Using this option will ignore return_record_name and
+        return_record_name_override
+    return_named_type_override
+        If true, this will modify the behavior of return_named_type so that
+        the named type is only returned for unions where there is more than
+        one named type. For unions that only have one named type, this option
+        will make it so that the named type is returned by itself, not a tuple
+        with the name
+    handle_unicode_errors
+        Default `strict`. Should be set to a valid string that can be used in
+        the errors argument of the string decode() function. Examples include
+        `replace` and `ignore`
 
 
     Example::
@@ -1035,14 +1026,32 @@ class reader(file_reader):
         The schema used when reading (if provided)
     """
 
-    def __init__(self, fo, reader_schema=None, return_record_name=False):
-        file_reader.__init__(self, fo, reader_schema, return_record_name)
+    def __init__(
+        self,
+        fo: Union[IO, AvroJSONDecoder],
+        reader_schema: Optional[Schema] = None,
+        return_record_name: bool = False,
+        return_record_name_override: bool = False,
+        handle_unicode_errors: str = "strict",
+        return_named_type: bool = False,
+        return_named_type_override: bool = False,
+    ):
+        options = {
+            "return_record_name": return_record_name,
+            "return_record_name_override": return_record_name_override,
+            "handle_unicode_errors": handle_unicode_errors,
+            "return_named_type": return_named_type,
+            "return_named_type_override": return_named_type_override,
+        }
+        super().__init__(fo, reader_schema, options)
 
         if isinstance(self.decoder, AvroJSONDecoder):
-            self.decoder.configure(self.reader_schema, self._named_schemas)
+            self.decoder.configure(self.reader_schema, self._named_schemas["reader"])
 
             self.writer_schema = self.reader_schema
             self.reader_schema = None
+            self._named_schemas["writer"] = self._named_schemas["reader"]
+            self._named_schemas["reader"] = {}
 
             def _elems():
                 while not self.decoder.done:
@@ -1051,7 +1060,7 @@ class reader(file_reader):
                         self.writer_schema,
                         self._named_schemas,
                         self.reader_schema,
-                        self.return_record_name,
+                        self.options,
                     )
                     self.decoder.drain()
 
@@ -1067,23 +1076,44 @@ class reader(file_reader):
                 self.writer_schema,
                 self._named_schemas,
                 self.reader_schema,
-                self.return_record_name,
+                self.options,
             )
 
 
-class block_reader(file_reader):
+class block_reader(file_reader[Block]):
     """Iterator over :class:`.Block` in an avro file.
 
     Parameters
     ----------
-    fo: file-like
+    fo
         Input stream
-    reader_schema: dict, optional
+    reader_schema
         Reader schema
-    return_record_name: bool, optional
+    return_record_name
         If true, when reading a union of records, the result will be a tuple
         where the first value is the name of the record and the second value is
         the record itself
+    return_record_name_override
+        If true, this will modify the behavior of return_record_name so that
+        the record name is only returned for unions where there is more than
+        one record. For unions that only have one record, this option will make
+        it so that the record is returned by itself, not a tuple with the name.
+    return_named_type
+        If true, when reading a union of named types, the result will be a tuple
+        where the first value is the name of the type and the second value is
+        the record itself
+        NOTE: Using this option will ignore return_record_name and
+        return_record_name_override
+    return_named_type_override
+        If true, this will modify the behavior of return_named_type so that
+        the named type is only returned for unions where there is more than
+        one named type. For unions that only have one named type, this option
+        will make it so that the named type is returned by itself, not a tuple
+        with the name
+    handle_unicode_errors
+        Default `strict`. Should be set to a valid string that can be used in
+        the errors argument of the string decode() function. Examples include
+        `replace` and `ignore`
 
 
     Example::
@@ -1111,8 +1141,24 @@ class block_reader(file_reader):
         The schema used when reading (if provided)
     """
 
-    def __init__(self, fo, reader_schema=None, return_record_name=False):
-        file_reader.__init__(self, fo, reader_schema, return_record_name)
+    def __init__(
+        self,
+        fo: IO,
+        reader_schema: Optional[Schema] = None,
+        return_record_name: bool = False,
+        return_record_name_override: bool = False,
+        handle_unicode_errors: str = "strict",
+        return_named_type: bool = False,
+        return_named_type_override: bool = False,
+    ):
+        options = {
+            "return_record_name": return_record_name,
+            "return_record_name_override": return_record_name_override,
+            "handle_unicode_errors": handle_unicode_errors,
+            "return_named_type": return_named_type,
+            "return_named_type_override": return_named_type_override,
+        }
+        super().__init__(fo, reader_schema, options)
 
         self._read_header()
 
@@ -1123,33 +1169,63 @@ class block_reader(file_reader):
             self.writer_schema,
             self._named_schemas,
             self.reader_schema,
-            self.return_record_name,
+            self.options,
         )
 
 
-def schemaless_reader(fo, writer_schema, reader_schema=None, return_record_name=False):
-    """Reads a single record writen using the
+def schemaless_reader(
+    fo: IO,
+    writer_schema: Schema,
+    reader_schema: Optional[Schema] = None,
+    return_record_name: bool = False,
+    return_record_name_override: bool = False,
+    handle_unicode_errors: str = "strict",
+    return_named_type: bool = False,
+    return_named_type_override: bool = False,
+) -> AvroMessage:
+    """Reads a single record written using the
     :meth:`~fastavro._write_py.schemaless_writer`
 
     Parameters
     ----------
-    fo: file-like
+    fo
         Input stream
-    writer_schema: dict
+    writer_schema
         Schema used when calling schemaless_writer
-    reader_schema: dict, optional
+    reader_schema
         If the schema has changed since being written then the new schema can
         be given to allow for schema migration
-    return_record_name: bool, optional
+    return_record_name
         If true, when reading a union of records, the result will be a tuple
         where the first value is the name of the record and the second value is
         the record itself
+    return_record_name_override
+        If true, this will modify the behavior of return_record_name so that
+        the record name is only returned for unions where there is more than
+        one record. For unions that only have one record, this option will make
+        it so that the record is returned by itself, not a tuple with the name.
+    return_named_type
+        If true, when reading a union of named types, the result will be a tuple
+        where the first value is the name of the type and the second value is
+        the record itself
+        NOTE: Using this option will ignore return_record_name and
+        return_record_name_override
+    return_named_type_override
+        If true, this will modify the behavior of return_named_type so that
+        the named type is only returned for unions where there is more than
+        one named type. For unions that only have one named type, this option
+        will make it so that the named type is returned by itself, not a tuple
+        with the name
+    handle_unicode_errors
+        Default `strict`. Should be set to a valid string that can be used in
+        the errors argument of the string decode() function. Examples include
+        `replace` and `ignore`
 
 
     Example::
 
         parsed_schema = fastavro.parse_schema(schema)
-        with open('file.avro', 'rb') as fp:
+        with open('file', 'rb') as fp:
             record = fastavro.schemaless_reader(fp, parsed_schema)
 
     Note: The ``schemaless_reader`` can only read a single record.
@@ -1158,31 +1234,45 @@ def schemaless_reader(fo, writer_schema, reader_schema=None, return_record_name=
         # No need for the reader schema if they are the same
         reader_schema = None
 
-    named_schemas = {}
-    writer_schema = parse_schema(writer_schema, _named_schemas=named_schemas)
+    named_schemas: Dict[str, NamedSchemas] = _default_named_schemas()
+    writer_schema = parse_schema(writer_schema, named_schemas["writer"])
 
     if reader_schema:
-        reader_schema = parse_schema(reader_schema)
+        reader_schema = parse_schema(reader_schema, named_schemas["reader"])
 
     decoder = BinaryDecoder(fo)
 
+    options = {
+        "return_record_name": return_record_name,
+        "return_record_name_override": return_record_name_override,
+        "handle_unicode_errors": handle_unicode_errors,
+        "return_named_type": return_named_type,
+        "return_named_type_override": return_named_type_override,
+    }
+
     return read_data(
         decoder,
         writer_schema,
         named_schemas,
         reader_schema,
-        return_record_name,
+        options,
     )
 
 
-def is_avro(path_or_buffer):
-    """Return True if path (or buffer) points to an Avro file.
+def is_avro(path_or_buffer: Union[str, IO]) -> bool:
+    """Return True if path (or buffer) points to an Avro file. This will only
+    work for avro files that contain the normal avro schema header like those
+    create from :func:`~fastavro._write_py.writer`. This function is not intended
+    to be used with binary data created from
+    :func:`~fastavro._write_py.schemaless_writer` since that does not include the
+    avro header.
 
     Parameters
     ----------
-    path_or_buffer: path to file or file-like object
+    path_or_buffer
         Path to file
     """
+    fp: IO
     if isinstance(path_or_buffer, str):
         fp = open(path_or_buffer, "rb")
         close = True
