@@ -13,8 +13,8 @@ if [[ "$SKIP_BLACK" = "1" ]]; then
 	true
 else
 	echo "running black"
-	black --target-version py36 --diff fastavro/ tests/ setup.py
-	black --target-version py36 --check fastavro/ tests/ setup.py
+	black --target-version py38 --diff fastavro/ tests/ setup.py
+	black --target-version py38 --check fastavro/ tests/ setup.py
 fi
 
 echo "running flake8"
@@ -42,7 +42,7 @@ fi
 check-manifest
 
 # Build Cython modules
-FASTAVRO_USE_CYTHON=1 python setup.py build_ext --inplace
+python setup.py build_ext --inplace
 pip install -e .
 
 PYTHONPATH=${PWD} python -m pytest --cov=fastavro -v --cov-report=term-missing --cov-report=html:build/htmlcov $@
