@@ -130,7 +130,8 @@ def test_unsupported_codec():
         list(fastavro.reader(modified_file))
 
 
-def test_compression_level():
+@pytest.mark.parametrize("codec", ["deflate", "zstandard"])
+def test_compression_level(codec):
     """https://github.com/fastavro/fastavro/issues/377"""
     schema = {
         "doc": "A weather reading.",
@@ -152,8 +153,22 @@ def test_compression_level():
     ]
 
     file = BytesIO()
-    fastavro.writer(file, schema, records, codec="deflate", codec_compression_level=9)
+    fastavro.writer(file, schema, records, codec=codec, codec_compression_level=9)
 
     file.seek(0)
     out_records = list(fastavro.reader(file))
     assert records == out_records
+
+
+def test_zstandard_decompress_stream():
+    """https://github.com/fastavro/fastavro/pull/575"""
+    binary = (
+        b'Obj\x01\x04\x14avro.codec\x12zstandard\x16avro.schema\xc6\x01{"name"'
+        + b':"Weather","namespace":"test","type":"record","fields":[{"name":"s'
+        + b'tation","type":"string"}]}\x001234567890123456\x02\x1c(\xb5/\xfd\x00'
+        + b"X)\x00\x00\x08AAAA1234567890123456"
+    )
+
+    file = BytesIO(binary)
+    out_records = list(fastavro.reader(file))
+    assert [{"station": "AAAA"}] == out_records
