@@ -1,6 +1,8 @@
+from io import BytesIO
 from os.path import join, abspath, dirname
 import pytest
 import fastavro
+from fastavro.repository import AbstractSchemaRepository
 from fastavro.schema import (
     SchemaParseException,
     UnknownType,
@@ -11,44 +13,22 @@ from fastavro.schema import (
 
 
 def test_named_types_have_names():
-    record_schema = {
-        "type": "record",
-        "fields": [
-            {
-                "name": "field",
-                "type": "string",
-            }
-        ],
-    }
+    record_schema = {"type": "record", "fields": [{"name": "field", "type": "string"}]}
 
     with pytest.raises(SchemaParseException):
         fastavro.parse_schema(record_schema)
 
-    error_schema = {
-        "type": "error",
-        "fields": [
-            {
-                "name": "field",
-                "type": "string",
-            }
-        ],
-    }
+    error_schema = {"type": "error", "fields": [{"name": "field", "type": "string"}]}
 
     with pytest.raises(SchemaParseException):
         fastavro.parse_schema(error_schema)
 
-    fixed_schema = {
-        "type": "fixed",
-        "size": 1,
-    }
+    fixed_schema = {"type": "fixed", "size": 1}
 
     with pytest.raises(SchemaParseException):
         fastavro.parse_schema(fixed_schema)
 
-    enum_schema = {
-        "type": "enum",
-        "symbols": ["FOO"],
-    }
+    enum_schema = {"type": "enum", "symbols": ["FOO"]}
 
     with pytest.raises(SchemaParseException):
         fastavro.parse_schema(enum_schema)
@@ -63,12 +43,7 @@ def test_parse_schema():
     schema = {
         "type": "record",
         "name": "test_parse_schema",
-        "fields": [
-            {
-                "name": "field",
-                "type": "string",
-            }
-        ],
+        "fields": [{"name": "field", "type": "string"}],
     }
 
     parsed_schema = parse_schema(schema)
@@ -79,25 +54,15 @@ def test_parse_schema():
 
 
 def test_unknown_type():
-    schema = {
-        "type": "unknown",
-    }
-
     with pytest.raises(UnknownType):
-        parse_schema(schema)
+        parse_schema({"type": "unknown"})
 
 
 def test_aliases_are_preserved():
     schema = {
         "type": "record",
         "name": "test_parse_schema",
-        "fields": [
-            {
-                "name": "field",
-                "type": "string",
-                "aliases": ["test"],
-            }
-        ],
+        "fields": [{"name": "field", "type": "string", "aliases": ["test"]}],
     }
 
     parsed_schema = parse_schema(schema)
@@ -109,13 +74,7 @@ def test_aliases_is_a_list():
     schema = {
         "type": "record",
         "name": "test_parse_schema",
-        "fields": [
-            {
-                "name": "field",
-                "type": "string",
-                "aliases": "foobar",
-            }
-        ],
+        "fields": [{"name": "field", "type": "string", "aliases": "foobar"}],
     }
 
     with pytest.raises(SchemaParseException):
@@ -141,7 +100,7 @@ def test_decimal_scale_is_an_int():
     }
 
     with pytest.raises(
-        SchemaParseException, match="decimal scale must be a postive integer"
+        SchemaParseException, match="decimal scale must be a positive integer"
     ):
         parse_schema(schema)
 
@@ -165,7 +124,7 @@ def test_decimal_scale_is_a_positive_int():
     }
 
     with pytest.raises(
-        SchemaParseException, match="decimal scale must be a postive integer"
+        SchemaParseException, match="decimal scale must be a positive integer"
     ):
         parse_schema(schema)
 
@@ -190,7 +149,7 @@ def test_decimal_precision_is_an_int():
 
     with pytest.raises(
         SchemaParseException,
-        match="decimal precision must be a postive integer",
+        match="decimal precision must be a positive integer",
     ):
         parse_schema(schema)
 
@@ -215,7 +174,7 @@ def test_decimal_precision_is_a_positive_int():
 
     with pytest.raises(
         SchemaParseException,
-        match="decimal precision must be a postive integer",
+        match="decimal precision must be a positive integer",
     ):
         parse_schema(schema)
 
@@ -278,19 +237,11 @@ def test_named_type_cannot_be_redefined():
         "namespace": "test.avro.training",
         "name": "SomeMessage",
         "fields": [
-            {
-                "name": "is_error",
-                "type": "boolean",
-                "default": False,
-            },
+            {"name": "is_error", "type": "boolean", "default": False},
             {
                 "name": "outcome",
                 "type": [
-                    {
-                        "type": "record",
-                        "name": "SomeMessage",
-                        "fields": [],
-                    },
+                    {"type": "record", "name": "SomeMessage", "fields": []},
                     {
                         "type": "record",
                         "name": "ErrorRecord",
@@ -319,19 +270,11 @@ def test_named_type_cannot_be_redefined():
         "fields": [
             {
                 "name": "field1",
-                "type": {
-                    "type": "record",
-                    "name": "ThisName",
-                    "fields": [],
-                },
+                "type": {"type": "record", "name": "ThisName", "fields": []},
             },
             {
                 "name": "field2",
-                "type": {
-                    "type": "enum",
-                    "name": "ThisName",
-                    "symbols": ["FOO", "BAR"],
-                },
+                "type": {"type": "enum", "name": "ThisName", "symbols": ["FOO", "BAR"]},
             },
         ],
     }
@@ -345,19 +288,11 @@ def test_named_type_cannot_be_redefined():
         "fields": [
             {
                 "name": "field1",
-                "type": {
-                    "type": "record",
-                    "name": "ThatName",
-                    "fields": [],
-                },
+                "type": {"type": "record", "name": "ThatName", "fields": []},
             },
             {
                 "name": "field2",
-                "type": {
-                    "type": "fixed",
-                    "name": "ThatName",
-                    "size": 8,
-                },
+                "type": {"type": "fixed", "name": "ThatName", "size": 8},
             },
         ],
     }
@@ -541,6 +476,27 @@ def test_union_schemas_must_have_names_in_order():
         parse_schema(schema2)
 
 
+def test_using_named_schemas_to_handle_references():
+    location = {
+        "name": "Location",
+        "type": "record",
+        "fields": [{"name": "city", "type": "long"}],
+    }
+    weather = {
+        "name": "Weather",
+        "type": "record",
+        "fields": [{"name": "of", "type": "Location"}],
+    }
+
+    named_schemas = {}
+    parse_schema(location, named_schemas)
+    parse_schema(weather, named_schemas)
+
+    # This should not work because didn't supply the named schemas
+    with pytest.raises(UnknownType):
+        parse_schema(weather)
+
+
 def test_load_schema_does_not_make_unions_of_unions():
     """https://github.com/fastavro/fastavro/issues/443"""
     load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test")
@@ -998,3 +954,404 @@ def test_load_schema_ordered():
         ],
     }
     assert loaded_schema == expected_schema
+
+
+def test_load_schema_top_level_names():
+    """https://github.com/fastavro/fastavro/issues/527"""
+    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test_13")
+    schema_path = join(load_schema_dir, "A.avsc")
+    loaded_schema = fastavro.schema.load_schema(schema_path, _write_hint=False)
+
+    expected_schema = {
+        "name": "B",
+        "type": "record",
+        "fields": [{"name": "foo", "type": "string"}],
+    }
+    assert loaded_schema == expected_schema
+
+
+def test_load_schema_top_level_primitive():
+    """https://github.com/fastavro/fastavro/issues/527"""
+    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test_14")
+    schema_path = join(load_schema_dir, "A.avsc")
+    loaded_schema = fastavro.schema.load_schema(schema_path, _write_hint=False)
+
+    expected_schema = "string"
+    assert loaded_schema == expected_schema
+
+
+def test_load_schema_union_names():
+    """https://github.com/fastavro/fastavro/issues/527"""
+    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test_15")
+    schema_path = join(load_schema_dir, "A.avsc")
+    loaded_schema = fastavro.schema.load_schema(schema_path, _write_hint=False)
+
+    expected_schema = [
+        {
+            "name": "B",
+            "type": "record",
+            "fields": [{"name": "foo", "type": "string"}],
+        },
+        {
+            "name": "C",
+            "type": "record",
+            "fields": [{"name": "bar", "type": "string"}],
+        },
+    ]
+    assert loaded_schema == expected_schema
+
+
+def test_load_schema_accepts_custom_repository():
+    class LocalSchemaRepository(AbstractSchemaRepository):
+        def __init__(self, schemas):
+            self.schemas = schemas
+
+        def load(self, subject):
+            return self.schemas.get(subject)
+
+    repo = LocalSchemaRepository(
+        {
+            "A": {
+                "name": "A",
+                "type": "record",
+                "fields": [{"name": "foo", "type": "B"}],
+            },
+            "B": {
+                "name": "B",
+                "type": "record",
+                "fields": [{"name": "bar", "type": "string"}],
+            },
+        }
+    )
+
+    loaded_schema = fastavro.schema.load_schema("A", repo=repo, _write_hint=False)
+
+    expected_schema = {
+        "name": "A",
+        "type": "record",
+        "fields": [
+            {
+                "name": "foo",
+                "type": {
+                    "name": "B",
+                    "type": "record",
+                    "fields": [{"name": "bar", "type": "string"}],
+                },
+            },
+        ],
+    }
+    assert loaded_schema == expected_schema
+
+
+def test_schema_expansion_3():
+    """https://github.com/fastavro/fastavro/issues/538"""
+    references = {
+        "com.namespace.dependencies.Dependency": {
+            "name": "Dependency",
+            "namespace": "com.namespace.dependencies",
+            "type": "record",
+            "fields": [{"name": "sub_field_1", "type": "string"}],
+        }
+    }
+
+    original_schema = {
+        "name": "MasterSchema",
+        "namespace": "com.namespace.master",
+        "type": "record",
+        "fields": [
+            {"name": "field_2", "type": "com.namespace.dependencies.Dependency"}
+        ],
+    }
+
+    expected_expanded_schema_fields = [
+        {
+            "name": "field_2",
+            "type": {
+                "name": "Dependency",
+                "namespace": "com.namespace.dependencies",
+                "type": "record",
+                "fields": [{"name": "sub_field_1", "type": "string"}],
+            },
+        }
+    ]
+
+    assert isinstance(original_schema, dict)
+
+    try:
+        parsed_schema = parse_schema(original_schema, named_schemas=references)
+        assert expected_expanded_schema_fields == expand_schema(parsed_schema)["fields"]
+    except UnknownType:
+        pytest.fail(
+            "expand_schema raised UnknownType even though referenced type is part of named_schemas"
+        )
+
+
+def test_explicit_null_namespace():
+    """https://github.com/fastavro/fastavro/issues/537"""
+    schema = {
+        "type": "record",
+        "name": "my_schema",
+        "namespace": "",
+        "fields": [{"name": "subfield", "type": "string"}],
+    }
+    parsed_schema = parse_schema(schema)
+    assert parsed_schema["name"] == "my_schema"
+
+
+def test_explicit_null_namespace_2():
+    """https://github.com/fastavro/fastavro/issues/537"""
+    schema = {
+        "type": "record",
+        "name": "my_schema",
+        "namespace": None,
+        "fields": [{"name": "subfield", "type": "string"}],
+    }
+    parsed_schema = parse_schema(schema)
+    assert parsed_schema["name"] == "my_schema"
+
+
+@pytest.mark.parametrize("symbol", [None, 0, "Ż", "0nope"])
+def test_enum_symbols_validation__invalid(symbol):
+    """https://github.com/fastavro/fastavro/issues/551"""
+    invalid_schema = {
+        "type": "record",
+        "name": "my_schema",
+        "fields": [
+            {
+                "name": "enum_field",
+                "type": {
+                    "name": "my_enum",
+                    "type": "enum",
+                    "symbols": [symbol],
+                },
+            }
+        ],
+    }
+
+    with pytest.raises(SchemaParseException) as err:
+        parse_schema(invalid_schema)
+
+    assert (
+        str(err.value)
+        == "Every symbol must match the regular expression [A-Za-z_][A-Za-z0-9_]*"
+    )
+
+
+@pytest.mark.parametrize("symbol", ["OK", "_123", "None", "Enum"])
+def test_enum_symbols_validation__correct(symbol):
+    """https://github.com/fastavro/fastavro/issues/551"""
+    invalid_schema = {
+        "type": "record",
+        "name": "my_schema",
+        "fields": [
+            {
+                "name": "enum_field",
+                "type": {
+                    "name": "my_enum",
+                    "type": "enum",
+                    "symbols": [symbol],
+                },
+            }
+        ],
+    }
+
+    try:
+        parse_schema(invalid_schema)
+    except SchemaParseException:
+        pytest.fail(f"valid symbol {symbol} has been incorrectly marked as invalid.")
+
+
+def test_enum_symbols_validation__uniqueness():
+    """https://github.com/fastavro/fastavro/issues/551"""
+    invalid_schema = {
+        "type": "record",
+        "name": "my_schema",
+        "fields": [
+            {
+                "name": "enum_field",
+                "type": {
+                    "name": "my_enum",
+                    "type": "enum",
+                    "symbols": ["FOO", "BAR", "FOO"],
+                },
+            }
+        ],
+    }
+
+    with pytest.raises(SchemaParseException) as err:
+        parse_schema(invalid_schema)
+
+    assert str(err.value) == "All symbols in an enum must be unique"
+
+
+def test_enum_with_bad_default():
+    """https://github.com/fastavro/fastavro/issues/563"""
+
+    schema = {
+        "type": "record",
+        "name": "test_schema",
+        "fields": [
+            {
+                "name": "test_enum",
+                "type": {
+                    "name": "test_enum_type",
+                    "type": "enum",
+                    "symbols": ["NONE"],
+                    "default": "UNKNOWN",
+                },
+            }
+        ],
+    }
+
+    with pytest.raises(
+        SchemaParseException, match="Default value for enum must be in symbols list"
+    ):
+        fastavro.parse_schema(schema)
+
+
+def test_foobar():
+    """https://github.com/fastavro/fastavro/issues/608"""
+    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test_16")
+    schema_path = join(load_schema_dir, "namespace.match.avsc")
+    loaded_schema = fastavro.schema.load_schema(schema_path)
+
+    record = {
+        "id": "123",
+        "team1": {"name": "A"},
+        "team2": {"name": "B"},
+    }
+
+    bio = BytesIO()
+    fastavro.writer(bio, loaded_schema, [record])
+    fastavro.writer(bio, loaded_schema, [record])
+
+    bio.seek(0)
+
+    records = list(fastavro.reader(bio))
+    assert records == [record, record]
+
+
+def test_default_matches_first_union_type():
+    """https://github.com/fastavro/fastavro/issues/649"""
+
+    schema = {
+        "type": "record",
+        "name": "test_default_matches_first_union_type",
+        "fields": [
+            {
+                "name": "first_union",
+                "type": ["string", "long"],
+                "default": 10,
+            }
+        ],
+    }
+
+    with pytest.raises(
+        SchemaParseException,
+        match="Default value <10> must match first schema in union",
+    ):
+        fastavro.parse_schema(schema)
+
+
+@pytest.mark.parametrize(
+    "field_type,default_value",
+    (
+        ("string", 0),
+        ({"type": "map", "values": "string"}, 0),
+        ({"type": "enum", "name": "somename", "symbols": ["FOO"]}, 0),
+        ({"type": "fixed", "name": "somename", "size": 4}, 0),
+        ({"type": "record", "name": "somename", "fields": []}, 0),
+        ({"type": "string"}, 0),
+    ),
+)
+def test_default_matches_type(field_type, default_value):
+    schema = {
+        "type": "record",
+        "name": "test_default_matches_type",
+        "fields": [
+            {
+                "name": "field",
+                "type": field_type,
+                "default": default_value,
+            }
+        ],
+    }
+
+    with pytest.raises(
+        SchemaParseException,
+        match=f"Default value <{default_value}> must match schema type",
+    ):
+        fastavro.parse_schema(schema)
+
+
+def test_correct_exception_text_for_default_value_test():
+    """https://github.com/fastavro/fastavro/issues/649"""
+
+    schema = {
+        "type": "record",
+        "name": "test_default_matches_first_union_type",
+        "fields": [
+            {
+                "name": "first_union",
+                "type": [
+                    "null",
+                    {
+                        "type": "record",
+                        "name": "internal",
+                        "fields": [
+                            {
+                                "name": "internal_field",
+                                "type": {"type": "array", "items": "string"},
+                                "default": 10,
+                            }
+                        ],
+                    },
+                ],
+                "default": None,
+            }
+        ],
+    }
+
+    with pytest.raises(
+        SchemaParseException,
+        match="Default value <10> must match schema type",
+    ):
+        fastavro.parse_schema(schema)
+
+
+@pytest.mark.parametrize("field_type,default_value", (("float", 0), ("double", 0)))
+def test_special_case_defaults(field_type, default_value):
+    schema = {
+        "type": "record",
+        "name": "test_special_case_defaults",
+        "fields": [
+            {
+                "name": "field",
+                "type": field_type,
+                "default": default_value,
+            }
+        ],
+    }
+
+    fastavro.parse_schema(schema)
+
+
+def test_namespace_respected():
+    """https://github.com/fastavro/fastavro/issues/690"""
+    schema = {
+        "type": "record",
+        "name": "explicit_namespace.foo",
+        "fields": [
+            {
+                "name": "field1",
+                "type": {
+                    "type": "record",
+                    "name": "bar",
+                    "fields": [{"name": "bar_field", "type": "int"}],
+                },
+            },
+            {"name": "field2", "type": "explicit_namespace.bar"},
+        ],
+    }
+
+    fastavro.parse_schema(schema)
