@@ -574,3 +574,42 @@ def test_record_name_with_named_type_in_union():
 
     parsed_schema = parse_schema(schema)
     validate_many(records, parsed_schema)
+
+
+def test_validate_should_not_parse_schema_if_it_was_parsed_already():
+    named_schemas = {}
+    parse_schema(
+        {"name": "B", "type": "record", "fields": [{"name": "bar", "type": "string"}]},
+        named_schemas,
+    )
+
+    a_schema = parse_schema(
+        {"name": "A", "type": "record", "fields": [{"name": "b", "type": "B"}]},
+        named_schemas,
+    )
+
+    records = [{"b": {"bar": "bar"}}]
+
+    validate_many(records, a_schema)
+    validate(records[0], a_schema)
+
+
+def test_validate_strict():
+    """https://github.com/fastavro/fastavro/issues/423"""
+    schema = {
+        "type": "record",
+        "name": "test_validate_strict",
+        "fields": [
+            {"name": "normal_field", "type": "string"},
+            {"name": "normal_union_field", "type": ["int", "float"]},
+            {"name": "nullable_union_field", "type": ["null", "string"]},
+        ],
+    }
+
+    record = {"normal_field": "a", "normal_union_field": 1}
+    parsed_schema = parse_schema(schema)
+    with pytest.raises(ValidationError):
+        validate(record, parsed_schema, strict=True)
+
+    with pytest.raises(ValidationError):
+        validate_many([record], parsed_schema, strict=True)
